package net.wanji.opt.service.impl;

import lombok.extern.slf4j.Slf4j;
import net.wanji.common.enums.ControlModeEnum;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.enums.TurnConvertEnum;
import net.wanji.common.enums.WeekEnum;
import net.wanji.common.framework.Constants;
import net.wanji.common.utils.tool.BeanListUtils;
import net.wanji.common.utils.tool.DateUtil;
import net.wanji.common.utils.tool.StringUtils;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.RidInfoMapper;
import net.wanji.feign.pojo.result.JsonViewObject;
import net.wanji.feign.pojo.vo.SchemeSendVO;
import net.wanji.feign.pojo.vo.SignalStatusVO;
import net.wanji.feign.service.UtcFeignClients;
import net.wanji.opt.dao.mapper.CrossPhaseMapper;
import net.wanji.opt.dao.mapper.CrossSchemeMapper;
import net.wanji.opt.dto.*;
import net.wanji.opt.po.base.CrossPhasePO;
import net.wanji.opt.po.base.CrossSchemePO;
import net.wanji.opt.service.CrossOptimizeService;
import net.wanji.opt.service.CrossSchedulesService;
import net.wanji.opt.service.CrossSchemeService;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author hfx
 * @date 2023/1/10 15:15
 * @desc CrossOptimizeServiceImpl
 */
@Service
@Slf4j
public class CrossOptimizeServiceImpl implements CrossOptimizeService {

    @Resource
    CrossSchemeService crossSchemeService;

    @Resource
    CrossSchedulesService crossSchedulesService;

    @Resource
    CrossPhaseMapper crossPhaseMapper;

    @Resource
    CrossSchemeMapper crossSchemeMapper;

    @Resource
    RidInfoMapper ridInfoMapper;

    @Resource
    UtcFeignClients utcFeignClients;

    static Set<String> CROSS_OPT = new HashSet<>();   // 记录已优化的路口

    @Override
    public String realtimeOptimize(List<CrossDataRealtimeDTO> abnormalCrossList) {

        try {
            // 相位配时信息，key：路口编号_方向类型_转向类型，value 相位配时信息
            Map<String, CrossPhaseDTO> phaseMap = listPhaseList();
            setTurnList(abnormalCrossList, phaseMap);

            // 获取所有信控路口转向实时数据
            Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtime = listTurnDataRealtime(abnormalCrossList);

            // 恢复优化路口原始方案
            restoreOptCrossOriSchema(abnormalCrossList, phaseMap);

            Map<Integer, List<CrossDataRealtimeDTO>> crossDataMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimeDTO::getStatus));
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.SPILLOVER.getCode());  // 溢出
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
//                spilloverOpt();
            }
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.CONGESTION.getCode()); // 拥堵
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
                congestionOpt(abnormalCrossList, turnDataRealtime, phaseMap);
            }
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.UNBALANCE.getCode());  // 失衡
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
                unbalanceOpt(abnormalCrossList, turnDataRealtime, phaseMap);
            }
        } catch (Exception e) {
            e.printStackTrace();
            log.error("实时优化", e.getMessage());
        }
        return null;
    }

    /**
     * 路口拥堵优化
     *
     * @param abnormalCrossList
     * @param turnDataRealtimeMap
     * @param phaseMap
     */
    private void congestionOpt(List<CrossDataRealtimeDTO> abnormalCrossList, Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap,
                               Map<String, CrossPhaseDTO> phaseMap) throws Exception {

        Map<String, List<CrossDataRealtimeDTO>> congestCrossMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimeDTO::getCrossId));

        String crossId;
        for (CrossDataRealtimeDTO cross : abnormalCrossList) {

            crossId = cross.getCrossId();
            // 判断信号机是否在线 todo 是否某些故障也不需要优化
            // 判断当前路口是否存在特殊控制操作
            if (isOffLineOrSpecialControlMode(crossId)) {
                continue;
            }

            // 获取上游路口
            List<RidInfoEntity> ridInfoEntities = ridInfoMapper.selectOne("", crossId);
            // 不存在上游路口，不做处理
            if (CollectionUtils.isEmpty(ridInfoEntities)) {
                continue;
            }

            // 判断当前路口是否已优化
            if (CROSS_OPT.contains(crossId)) {
                continue;
            }

            // 拥堵路口对上游路口进行优化
            for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
                // 拥堵列表包含上游路口，不需要操作, 否则当前路口进行优化 优化当前开始路口的此方向直行，和逆时针方向左转
                if (!congestCrossMap.containsKey(ridInfoEntity.getEndCrossId())) {
                    String startCrossId = ridInfoEntity.getStartCrossId();
                    doExecuteCrossOpt(turnDataRealtimeMap, phaseMap, startCrossId);
                }
            }
            // 判断路口是否是绿波

            // 记录已优化的路口
            CROSS_OPT.add(crossId);
        }
    }

    /**
     * 设置异常数据转向数据
     *
     * @param abnormalCrossList
     * @param phaseMap
     */
    private static void setTurnList(List<CrossDataRealtimeDTO> abnormalCrossList, Map<String, CrossPhaseDTO> phaseMap) {
        for (CrossDataRealtimeDTO crossDataRealtimeDTO : abnormalCrossList) {
            List<CrossTurnInfoDTO> turnList = new ArrayList<>();
            for (Map.Entry<String, CrossPhaseDTO> entry : phaseMap.entrySet()) {
                String crossDirTurn = entry.getKey();
                String[] split = crossDirTurn.split(Constants.SystemParam.SEPARATOR_UNDER_LINE);
                String crossId = split[0];
                String dir = split[1];
                String turn = split[2];

                CrossTurnInfoDTO crossTurnInfoDTO = new CrossTurnInfoDTO();
                if (StringUtils.equalsIgnoreCase(crossId, crossDataRealtimeDTO.getCrossId())) {
                    crossTurnInfoDTO.setTurnType(Integer.valueOf(turn));
                    crossTurnInfoDTO.setInDir(Integer.valueOf(dir));
                    turnList.add(crossTurnInfoDTO);
                }
            }
            crossDataRealtimeDTO.setTurnList(turnList);
        }
    }

    /**
     * 通过第二批次列表，判断需要恢复的优化路口原始方案
     *
     * @param abnormalCrossList
     */
    private void restoreOptCrossOriSchema(List<CrossDataRealtimeDTO> abnormalCrossList, Map<String, CrossPhaseDTO> phaseMap) throws Exception {
        List<String> abnormalCrossIdList = abnormalCrossList.stream().map(CrossDataRealtimeDTO::getCrossId).collect(Collectors.toList());
        for (String optCrossId : CROSS_OPT) {
            if (!abnormalCrossIdList.contains(optCrossId)) {
                // 下发原始方案
                SchemeSendVO schemeSendVO = getSchemeSendVO(optCrossId, phaseMap, null);
                JsonViewObject jsonViewObject = utcFeignClients.schemeSend(schemeSendVO);
                if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
                    log.error("实时监控路口优化方案下发远程调用异常！", jsonViewObject);
                    throw new Exception();
                }
            }
        }
    }

    /**
     * 路口失衡优化
     *
     * @param abnormalCrossList   失衡路口实时数据
     * @param turnDataRealtimeMap 路口转向实时数据
     * @param phaseMap            路口相位配时数据
     */
    private void unbalanceOpt(List<CrossDataRealtimeDTO> abnormalCrossList, Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap,
                              Map<String, CrossPhaseDTO> phaseMap) throws Exception {

        String crossId;
        for (CrossDataRealtimeDTO cross : abnormalCrossList) {

            crossId = cross.getCrossId();
            // 判断信号机是否在线 todo 是否某些故障也不需要优化
            // 判断当前路口是否存在特殊控制操作
            if (isOffLineOrSpecialControlMode(crossId)) {
                continue;
            }
            // 判断当前路口是否已优化
            if (CROSS_OPT.contains(crossId)) {
                continue;
            }
            // 判断路口是否是绿波

            // 执行路口优化
            doExecuteCrossOpt(turnDataRealtimeMap, phaseMap, crossId);

            // 记录已优化的路口
            CROSS_OPT.add(crossId);
        }

    }

    /**
     * 执行路口优化逻辑
     * 计算路口相位优化时长并下发方案
     *
     * @param turnDataRealtimeMap
     * @param phaseMap
     * @param crossId
     * @throws Exception
     */
    private void doExecuteCrossOpt(Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap, Map<String, CrossPhaseDTO> phaseMap, String crossId) throws Exception {
        // 计算路口各转向所需的通行时长（秒）
        List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = turnDataRealtimeMap.get(crossId);
        crossTurnDataRealtimeDTOS.forEach(item -> item.setPassTime(item.calPassTime(item)));

        // 计算路口各相位调整时长（秒）
        Map<String, Integer> phaseTimeOptResultMap = new HashMap<>();
        Map<Integer, String> timeOffsetPhaseMap = getTimeOffsetPhaseMap(phaseMap, crossTurnDataRealtimeDTOS);
        // 是否需要拆分相位
        if (isSplitPhase(timeOffsetPhaseMap)) {

        } else {
            phaseTimeOptResultMap = getPhaseTimeOptMap(timeOffsetPhaseMap);
        }
        // 修改原始方案绿灯时间为优化后时间，方案下发
        SchemeSendVO schemeSendVO = getSchemeSendVO(crossId, phaseMap, phaseTimeOptResultMap);
        JsonViewObject jsonViewObject = utcFeignClients.schemeSend(schemeSendVO);
        if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
            log.error("实时监控路口优化方案下发远程调用异常！", jsonViewObject);
            throw new Exception();
        }
    }

    /**
     * 构建下发的优化方案
     * 如果phaseTimeOptResultMap=null,下发的原始方案
     * 否则原始方案和相位优化时间参数phaseTimeOptResultMap构建下发参数
     *
     * @param crossId
     * @param phaseMap
     * @param phaseTimeOptResultMap
     * @return
     */
    private SchemeSendVO getSchemeSendVO(String crossId, Map<String, CrossPhaseDTO> phaseMap, Map<String, Integer> phaseTimeOptResultMap) {
        String key = String.join(Constants.SystemParam.SEPARATOR_UNDER_LINE, "1", "s");
        CrossPhaseDTO crossPhaseDTO = phaseMap.get(key);
        String planNo = crossPhaseDTO.getPlanNo();

        List<CrossSchemePO> crossSchemePOS = crossSchemeMapper.listCrossSchemeInfo(crossId, planNo);
        List<CrossPhasePO> crossPhasePOS = crossPhaseMapper.listCrossPhase(crossId, planNo);
        Map<Integer, List<CrossPhasePO>> ringNoPhasePOMap = crossPhasePOS.stream().collect(Collectors.groupingBy(CrossPhasePO::getRingNo));

        SchemeSendVO schemeSendVO = new SchemeSendVO();
        List<SchemeSendVO.Pattern> patternList = new ArrayList<>();
        BeanListUtils.populateList(crossSchemePOS, patternList, SchemeSendVO.Pattern.class);
        schemeSendVO.setPatternList(patternList);
        schemeSendVO.setCrossCode(crossId);

        List<SchemeSendVO.Pattern.Ring> rings = new ArrayList<>();
        ringNoPhasePOMap.entrySet().forEach(entry -> {
            SchemeSendVO.Pattern.Ring ring = new SchemeSendVO.Pattern.Ring();
            List<SchemeSendVO.Pattern.Ring.Phase> phaseList = new ArrayList<>();
            List<CrossPhasePO> value = entry.getValue();
            BeanListUtils.populateList(value, phaseList, SchemeSendVO.Pattern.Ring.Phase.class);
            ring.setRingNo(String.valueOf(entry.getKey()));
            // 在原始方案基础绿灯时间+可优化时间
            if (!CollectionUtils.isEmpty(phaseTimeOptResultMap)) {
                phaseList.forEach(phaseInfo -> {
                    Integer optTime = phaseTimeOptResultMap.get(phaseInfo.getPhaseNo());
                    String greenTime = phaseInfo.getGreenTime();
                    phaseInfo.setGreenTime(greenTime + optTime);
                });
            }
            ring.setPhaseList(phaseList);
            rings.add(ring);
        });

        schemeSendVO.setCrossCode(crossId);
        schemeSendVO.setPatternList(patternList);
        return schemeSendVO;
    }

    /**
     * 在最大最小绿范围内，通过转向优化计算相位可优化时间
     *
     * @param timeOffsetPhaseMap
     * @return
     */
    private static Map<String, Integer> getPhaseTimeOptMap(Map<Integer, String> timeOffsetPhaseMap) {
        HashMap<String, Integer> phaseTimeOptResultMap = new HashMap<>();
        // key:相位号 value:最大可加减时间
        Map<String, Integer> phaseOffsetTimeSetMap = new HashMap<>();
        for (Map.Entry<Integer, String> entry : timeOffsetPhaseMap.entrySet()) {
            Integer offsetTime = entry.getKey();
            String phaseNo = entry.getValue();
            if (timeOffsetPhaseMap.containsValue(phaseNo)) {
                Integer currentOffsetTime = phaseOffsetTimeSetMap.get(phaseNo);
                if (Math.abs(offsetTime) > Math.abs(currentOffsetTime)) {
                    phaseOffsetTimeSetMap.put(phaseNo, offsetTime);
                } else {
                    phaseOffsetTimeSetMap.put(phaseNo, currentOffsetTime);
                }
            } else {
                phaseOffsetTimeSetMap.put(phaseNo, offsetTime);
            }
        }
        List<Integer> subTimeList = phaseOffsetTimeSetMap.entrySet().stream().map(entry -> entry.getValue()).collect(Collectors.toList()).stream().filter(i -> i < 0).collect(Collectors.toList());
        int subTimeSum = subTimeList.stream().mapToInt(Integer::intValue).sum();
        int lastAddTimeSum = 0;
        for (Map.Entry<String, Integer> entry : phaseOffsetTimeSetMap.entrySet()) {
            String phaseNo = entry.getKey();
            Integer offsetTime = entry.getValue();
            if (offsetTime > 0 && subTimeSum + offsetTime <= 0) {
                // 当前相位可加时间就是最大可加时间，可加时间有可减时间
                subTimeSum += offsetTime;
                lastAddTimeSum += offsetTime;
                phaseTimeOptResultMap.put(phaseNo, offsetTime);
            }
            if (offsetTime < 0) {
                if (lastAddTimeSum <= Math.abs(offsetTime)) {
                    phaseTimeOptResultMap.put(phaseNo, lastAddTimeSum < 0 ? lastAddTimeSum : -lastAddTimeSum);
                } else {
                    lastAddTimeSum += offsetTime;
                    phaseTimeOptResultMap.put(phaseNo, offsetTime);
                }
            }
        }
        return phaseTimeOptResultMap;
    }

    /**
     * 是否需要拆分相位灯组
     * 同一相位内不同转向同时存在可加可减，需要拆分相位
     *
     * @param timeOffsetPhaseMap
     * @return
     */
    private boolean isSplitPhase(Map<Integer, String> timeOffsetPhaseMap) {
        HashMap<String, Boolean> phaseOffsetMap = new HashMap<>();
        for (Map.Entry<Integer, String> entry : timeOffsetPhaseMap.entrySet()) {
            Boolean gtZore = entry.getKey() >= 0 ? Boolean.TRUE : Boolean.FALSE;
            String phaseNo = entry.getValue();
            if (phaseOffsetMap.containsKey(phaseNo) && !phaseOffsetMap.containsValue(gtZore)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 计算路口各相位调整时长（秒）
     * 相位所需总放行时长 = sum(相位所需放行时长) - 总放行时长（周期）
     * 相位可减时长 = 相位有效绿灯时长 - 相位所需时长，同时“相位剩余有效绿灯时长”必须介于最小绿时长与最大绿时长之间
     * 相位可加时长 = 相位所需时长 - 有效绿灯时长
     * 相位可用总时长 = sum(相位可减时长)
     * 相位优化时长 = 原相位时长 + 相位调整时长（相位可减时长或相位可加时长）
     *
     * @param phaseMap
     * @param crossTurnDataRealtimeDTOS
     * @return key:相位可加减时间 value:相位编号
     */
    private Map<Integer, String> getTimeOffsetPhaseMap(Map<String, CrossPhaseDTO> phaseMap, List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS) {
        Map<Integer, String> timeOffsetPhaseMap = new HashMap<>();
        crossTurnDataRealtimeDTOS.forEach(item -> {
            String key = String.join(Constants.SystemParam.SEPARATOR_UNDER_LINE, String.valueOf(item.getInDir()), String.valueOf(item.getTurnType()));
            Double passTime = item.getPassTime();

            phaseMap.entrySet().forEach(entry -> {
                String crossIdDirTurn = entry.getKey();
                CrossPhaseDTO crossPhaseDTO = entry.getValue();
                String phaseNo = crossPhaseDTO.getPhaseNo();

                if (StringUtils.equalsIgnoreCase(key, crossIdDirTurn)) {
                    Integer realOptGreenTime = getRealOptGreenTime(passTime, crossPhaseDTO);
                    Integer greenTimeOffset = realOptGreenTime - crossPhaseDTO.getGreenTime() - crossPhaseDTO.getGreenFlashTime();
                    timeOffsetPhaseMap.put(greenTimeOffset, phaseNo);
                }
            });
        });
        return timeOffsetPhaseMap;
    }

    private Integer getRealOptGreenTime(Double passTime, CrossPhaseDTO crossPhaseDTO) {
        int passTimeInt = passTime.intValue();
        if (passTimeInt <= crossPhaseDTO.getMinGreenTime()) {
            passTimeInt = crossPhaseDTO.getMinGreenTime();
        }
        if (passTime >= crossPhaseDTO.getMaxGreenTime()) {
            passTimeInt = crossPhaseDTO.getMaxGreenTime();
        }
        return passTimeInt;
    }


    private boolean isOffLineOrSpecialControlMode(String crossId) throws Exception {
        JsonViewObject jsonViewObject = utcFeignClients.runningStatusAlarm();
        if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
            log.error("路口优化获取当前路口控制模式异常！");
            throw new Exception();
        }
        List<SignalStatusVO> signalStatusVOS = (List<SignalStatusVO>) jsonViewObject.getContent();
        List<SignalStatusVO> resultList = signalStatusVOS.stream().filter(signalStatusVO -> Objects.equals(crossId, signalStatusVO.getCrossId())).collect(Collectors.toList());
        if (!CollectionUtils.isEmpty(resultList)) {
            SignalStatusVO signalStatusVO = resultList.get(0);
            if (0 == signalStatusVO.getStatus() || 1 != signalStatusVO.getControlType()) {
                return true;
            }
        }
        return false;
    }

    /**
     * 相位配时信息
     * key：路口编号_方向类型_转向类型
     * value：相位配时信息
     *
     * @return
     */
    private Map<String, CrossPhaseDTO> listPhaseList() throws Exception {

        Map<String, CrossPhaseDTO> phaseMap = new HashMap<>();
        // 获取已执行的调度详情信息
        List<CrossSchedulesDTO> schedulesDTOList = crossSchedulesService.listCrossSchedulesDetails(Constants.SystemParam.NULL, Constants.SystemParam.YES);
        // 获取当前运行时段信息
        List<CrossSectionDTO> sectionInfos = listSection(schedulesDTOList);
        // 获取方案详情信息
        List<CrossSchemeDTO> schemeDTOList = crossSchemeService.listCrossSchemeDetails(Constants.SystemParam.NULL);
        Map<Integer, CrossSchemeDTO> schemeMap = schemeDTOList.stream().collect(Collectors.toMap(CrossSchemeDTO::getId, e -> e));

        String key;
        String crossId;
        Integer dir;
        String[] turnArr;
        Set<String> turnSet = new HashSet<>();
        for (CrossSectionDTO section : sectionInfos) {                       // 运行时段列表

            crossId = section.getCrossId();
            CrossSchemeDTO scheme = schemeMap.get(section.getSchemeId());   // 方案信息
            for (CrossPhaseDTO phase : scheme.getPhaseInfos()) {             // 相位信息列表

                for (CrossLightsDTO lights : phase.getLightsInfos()) {       // 灯组信息列表

                    dir = lights.getDir();
                    for (LaneInfoDTO lane : lights.getLaneInfos()) {
                        // 车道转向转换为转向类型，并去重
                        turnArr = TurnConvertEnum.getCodeByKey(lane.getTurn()).split(Constants.SystemParam.SEPARATOR_UNDER_LINE);
                        turnSet.addAll(Arrays.stream(turnArr).collect(Collectors.toSet()));
                    }
                    for (String turn : turnSet) {

                        key = crossId + Constants.SystemParam.SEPARATOR_UNDER_LINE + dir + Constants.SystemParam.SEPARATOR_UNDER_LINE + turn;
                        phase.setDirType(dir);
                        phase.setTurnType(turn);
                        phaseMap.put(key, phase);
                    }
                    turnSet.clear();
                }
            }
        }
        return phaseMap;
    }

    /**
     * 获取当前运行时段信息
     *
     * @param schedulesDTOList 调度列表
     * @return
     */
    private List<CrossSectionDTO> listSection(List<CrossSchedulesDTO> schedulesDTOList) throws Exception {

        List<CrossSectionDTO> sectionInfos = new ArrayList<>();
        long currentTime = DateUtil.dateToStamp(DateUtil.getTime(), net.wanji.common.framework.Constants.DATE_FORMAT.E_DATE_FORMAT_TIME.getStrFormat());        // 获取当前时间
        int week = DateUtil.getWeek(new Date()) == Constants.SystemParam.ZERO ? WeekEnum.SUNDAY.getCode() : DateUtil.getWeek(new Date());   // 获取当前星期
        for (CrossSchedulesDTO schedules : schedulesDTOList) {

            if (Objects.equals(schedules.getWeek(), Constants.SystemParam.ZERO) &&
                    !DateUtil.getDate().equals(schedules.getSpecialDate())) {   // 先判断是否为特殊日期，如果特殊日期与当前日期不相同，则跳过
                continue;
            } else if (schedules.getWeek() != week) {
                continue;
            }

            // 时段信息列表
            for (CrossSectionDTO section : schedules.getSectionInfos()) {

                // 过滤非当前时段数据
                if (currentTime < DateUtil.dateToStamp(section.getStartTime()) || currentTime > DateUtil.dateToStamp(section.getEndTime())) {
                    continue;
                }
                // 过滤控制模式为“非定周期”的数据
                if (!ControlModeEnum.FIXED_PERIOD.equals(section.getControlMode())) {
                    continue;
                }
                section.setPlanNo(schedules.getPlanNo());
                sectionInfos.add(section);
            }
        }
        return sectionInfos;
    }


    /**
     * 获取所有信控路口转向实时数据
     * key 路口编号
     * value 转向实时数据
     *
     * @return
     */
    public Map<String, List<CrossTurnDataRealtimeDTO>> listTurnDataRealtime(List<CrossDataRealtimeDTO> abnormalCrossList) {
        Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeDTOList = new HashMap<>();

        Map<String, List<CrossDataRealtimeDTO>> abnormalCrossGroupMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimeDTO::getCrossId));
        abnormalCrossGroupMap.entrySet().forEach(entry -> {
            String crossId = entry.getKey();
            List<CrossDataRealtimeDTO> realTimeData = entry.getValue();
            List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = new ArrayList<>();
            realTimeData.forEach(realData -> {

                List<CrossTurnInfoDTO> turnList = realData.getTurnList();
                turnList.forEach(crossTurnInfoDTO -> {
                    CrossTurnDataRealtimeDTO crossTurnDataRealtimeDTO = new CrossTurnDataRealtimeDTO();
                    BeanUtils.copyProperties(realData, crossTurnDataRealtimeDTO);
                    crossTurnDataRealtimeDTO.setTurnType(crossTurnInfoDTO.getTurnType());
                    crossTurnDataRealtimeDTO.setInDir(crossTurnInfoDTO.getInDir());
                    crossTurnDataRealtimeDTO.setOutDir(crossTurnInfoDTO.getOutDir());
                    crossTurnDataRealtimeDTOS.add(crossTurnDataRealtimeDTO);
                });
            });
            turnDataRealtimeDTOList.put(crossId, crossTurnDataRealtimeDTOS);
        });

        return turnDataRealtimeDTOList;
    }

}
